#!/usr/local/bin/perl -T
#########
# Author:     Magnus Manske (mm6@sanger.ac.uk)
# Group:      Team 112
#

package settings ;

require Exporter;
use warnings ;
use strict ;
use AutoLoader qw(AUTOLOAD);
our @ISA = qw(Exporter);
our @EXPORT = qw( 
	$execpath $datapath $genome_file $webroot $docroot $htmlpath $tooltitle $reference_fa
	$cgi_path $annotation_file $snp_file $robots_flag $use_db_index_cutoff $uniqueness_file 
	$use_mysql $mysql_server $mysql_port $mysql_user $mysql_password @mysql_dbs
	%lscolor $ls_max_lines
	$bam_ftp $reference_ftp $render_image_command
	$css_file
	$paranoia_mode
);

# Tool title
our $tooltitle = 'LookSeq' ;

# The local file path to the CGI files
our $execpath = "/home/xianggan/bin/";

# The local file path to the directory containing the data (alignment databases, reference sequence etc.)
our $datapath = "/data/mus/bams/inbred/";

# The local file path to the HTML/JavaScript/i18n files
our $htmlpath = "/data/www/mus/";

# The full URL of the directory containing the HTML/JavaScript/i18n files
our $webroot = "http://mus.well.ox.ac.uk/mus/";

# The full URL of the directory containing the Perl files (this one, for example)
our $cgi_path = 'http://mus.well.ox.ac.uk/mus/cgi';

# The file name of the reference sequence in FASTA format. The file must be in the $datapath directory
our $genome_file = 'chr_all.fa' ;

# The file name of the reference sequence in SAMTOOLS .fa/.fai format. Give the .fa file only
our $reference_fa= 'chr_all.fa' ;

# The file name of the uniqueness (like FASTA, but each "base" is a digit 0-9, with 0=unique and 9=uh-oh). The file must be in the $datapath directory
our $uniqueness_file = '3D7_pm.fastu' ;

# Internal index usage cutoff; leave unless you know what you're doing!
our $use_db_index_cutoff = 25000 ;

# The file name of the sequence annotation database. The file must be in the $datapath directory.
# Leave undefined if no annotation is given.
###our $annotation_file = 'annotation.sqlite' ;

# The file name of the known SNP database. The file must be in the $datapath directory
# Leave undefined if no known SNP data is present.
###our $snp_file = "snps.sqlite" ;

# MySQL settings, deactivated by default
our $use_mysql = 0 ;
our ( $mysql_server , $mysql_port , $mysql_user , $mysql_password , @mysql_dbs ) ;


# Allows/disallows indexing of the page by robots (set to 'noindex' if you don't want the page to show up on Google et al)
our $robots_flag = 'index, follow' ;

# Don't use index for database query if region is larger than this
$use_db_index_cutoff = 20000 ;

# Colors
our %lscolor ;
$lscolor{'black'} = [ 0x00 , 0x00 , 0x00 ] ;
$lscolor{'white'} = [ 0xFF , 0xFF , 0xFF ] ;
$lscolor{'red'} = [ 0xFF , 0x00 , 0x00 ] ;
$lscolor{'blue'} = [ 0x00 , 0x00 , 0xFF ] ;
$lscolor{'ltblue'} = [ 0x44 , 0xB4 , 0xD5 ] ;
$lscolor{'green'} = [ 0x00 , 0xFF , 0x00 ] ;
$lscolor{'green2'} = [ 0x00 , 0x80 , 0x00 ] ;
$lscolor{'ltgreen'} = [ 0x72 , 0xFE , 0x95 ] ;
$lscolor{'yellow'} = [ 0xED , 0xEF , 0x85 ] ;
$lscolor{'ltgrey'} = [ 0xDC , 0xDC , 0xDC ] ;
$lscolor{'grey'} = [ 0xC8 , 0xC8 , 0xC8 ] ;
$lscolor{'brown'} = [ 0xD2 , 0x69 , 0x1E ] ; # Inversion
$lscolor{'cyan'} = [ 0x44 , 0xB4 , 0xD5 ] ; # Cigar
$lscolor{'mapqual_zero'} = [0xff, 0x72, 0x00] ;   # mapping quality zero

# BAM FTP
our $bam_ftp ;
our $reference_ftp ;

# BAM C++ render command, if available
our $render_image_command ;

# Stuff
our $css_file ;

# Upper SNP display limit for SAM format; set for max number of lines, e.g. 50000
our $ls_max_lines ;

our $paranoia_mode = 0 ;

1 ;
